import warnings
from datetime import date, datetime

import numpy as np

from astartes.utils.exceptions import InvalidConfigurationError, UncastableInputError
from astartes.utils.warnings import ConversionWarning


def convert_to_array(obj: object, name: str):
    """Attempt to convert obj named name to a numpy array, with appropriate warnings and exceptions.

    Args:
        obj (object): The item to attempt to convert.
        name (str): Human-readable name for printing.
    """
    warnings.warn(
        "Attempting to cast {:s} to a numpy array, which may result in unexpected behavior "
        "(remove this warning by passing numpy arrays directly to astartes).".format(
            name,
        ),
        ConversionWarning,
    )
    try:
        new_array = np.asarray(obj)
    except Exception as e:
        raise UncastableInputError(
            "Unable to cast {:s} to a numpy array using np.asarray."
        ) from e
    # ensure that all the values in the array are floats or date/datetime objects
    for item in new_array.ravel():
        if type(item) not in (float, np.float_, int, np.int_, date, datetime):
            raise InvalidConfigurationError(
                "After casting, input object {:s} contained unsupported type {:s}".format(
                    name,
                    str(type(item)),
                )
            )

    return new_array



import numpy as np
from scipy.spatial.distance import pdist, squareform

from astartes.samplers import AbstractSampler


class KennardStone(AbstractSampler):
    def __init__(self, *args):
        super().__init__(*args)

    def _sample(self):
        """
        Implements the Kennard-Stone algorithm
        """
        n_samples = len(self.X)

        distance_metric = self.get_config("metric", "euclidean")
        X_dist = pdist(self.X, metric=distance_metric)
        ks_distance = squareform(X_dist)
        # when searching for max distance, disregard self
        np.fill_diagonal(ks_distance, -np.inf)

        # get the row/col of maximum (greatest distance)
        max_idx = np.nanargmax(ks_distance)
        max_coords = np.unravel_index(max_idx, ks_distance.shape)

        # list of indices which have been selected
        # - used to mask ks_distance
        # - also tracks order of kennard-stone selection
        already_selected = list(max_coords)

        # minimum distance of all unselected samples to the two selected samples
        min_distances = np.min(ks_distance[:, already_selected], axis=1)
        for _ in range(n_samples - 2):
            # find the next sample with the largest minimum distance to any sample already selected
            already_selected.append(
                np.argmax(min_distances),
            )
            # get minimum distance of unselected samples to that sample only
            new_distances = np.min(ks_distance[:, [already_selected[-1]]], axis=1)
            min_distances = np.minimum(min_distances, new_distances)

        self._samples_idxs = np.array(already_selected, dtype=int)
